# -*- coding: utf-8 -*-

import torch
import math
import numpy as np
import torch.nn as nn
import sys
from torch.utils.checkpoint import checkpoint
from torch.nn.functional import interpolate
import torch.backends.cudnn as cudnn
from collections import OrderedDict
import cv2

__all__ = ['MyModel']

BN_f = None
BN_q = None

def remove_prefix(state_dict, prefix):
    f = lambda x: x.split(prefix, 1)[-1] if x.startswith(prefix) else x
    return {f(key): value for key, value in state_dict.items()}

class get_fc_E(nn.Module):
    def __init__(self, in_feature, in_h, in_w, out_feature):
        super(get_fc_E, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_feature, affine=True, eps=2e-5, momentum=0.9)
        self.dropout = nn.Dropout2d(p=0.4, inplace=True)
        self.fc = nn.Linear(in_feature*in_h*in_w, out_feature)
        self.bn2 = nn.BatchNorm1d(out_feature, affine=False, eps=2e-5, momentum=0.9)
    def forward(self, x):
        output = {}
        x = self.bn1(x)
        x = self.dropout(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        output['feature_nobn'] = x
        x = self.bn2(x)
        output['feature'] = x
        return output

class get_fc_EMom(nn.Module):
    def __init__(self, in_feature, in_h, in_w, out_feature):
        super().__init__()
        self.bn1 = nn.BatchNorm2d(in_feature, affine=True)
        # self.dropout = nn.Dropout2d(p=0.4, inplace=True)
        self.fc = nn.Linear(in_feature*in_h*in_w, out_feature)
        #TODO here momentum should be 0.1!
        self.bn2 = nn.BatchNorm1d(out_feature, affine=False, eps=2e-5, momentum=0.9, track_running_stats=True)
    def forward(self, x):
        output = {}
        x = self.bn1(x)
        # x = self.dropout(x)
        x = x.view(x.size(0), -1)
        x = self.fc(x)
        output['feature_nobn'] = x
        x = torch.cat((x, x), dim=0)
        x = self.bn2(x)
        output['feature'] = x[:1, :]
        return output

def conv1x1(in_planes, out_planes, stride=1):
    '''3x3 convolution with padding'''
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, padding=0, bias=False)

def conv3x3(in_planes, out_planes, stride=1, groups=1):
    '''3x3 convolution with padding'''
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride, padding=1, bias=False, groups=groups)

def activation(act_type='prelu'):
    if act_type == 'prelu':
        act = nn.PReLU()
    else:
        act = nn.ReLU(inplace=True)
    return act

class BasicBlock_v3(nn.Module):
    expansion = 1
    def __init__(self, inplanes, planes, stride=1, downsample=None, act_type='prelu', use_se=False, use_checkpoint=False):
        super(BasicBlock_v3, self).__init__()
        m = OrderedDict()
        m['bn1'] = BN_q(inplanes)
        m['conv1'] = conv3x3(inplanes, planes, stride=1)
        m['bn2'] = BN_q(planes)
        m['act1'] = activation(act_type)
        m['conv2'] = conv3x3(planes, planes, stride=stride)
        m['bn3'] = BN_q(planes)
        self.group1 = nn.Sequential(m)

        self.use_checkpoint = use_checkpoint
        self.use_se = use_se
        if self.use_se:
            s = OrderedDict()
            s['conv1'] = nn.Conv2d(planes, planes // 16, kernel_size=1, stride=1, padding=0)
            s['act1'] = activation(act_type)
            s['conv2'] = nn.Conv2d(planes // 16, planes, kernel_size=1, stride=1, padding=0)
            s['act2'] = nn.Sigmoid()
            self.se_block = nn.Sequential(s)

        self.act2 = activation(act_type)
        self.downsample = downsample

        bypass_bn_weight_list.append(self.group1.bn3.weight)

    def forward(self, x):
        if self.downsample is not None:
            residual = self.downsample(x)
        else:
            residual = x

        if self.use_checkpoint:
            out = checkpoint(self.group1, x)
        else:
            out = self.group1(x)

        if self.use_se:
            weight = F.adaptive_avg_pool2d(out, output_size=1)
            weight = self.se_block(weight)
            out = out * weight

        out = out + residual

        return self.act2(out)

class BasicConv2d(nn.Module):

    def __init__(self, in_channels, out_channels, kernel_size, stride, padding):
        super(BasicConv2d,self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size = kernel_size, stride = stride, padding = padding, bias = False)
        self.bn = BN_f(out_channels)
        self.relu = nn.ReLU(inplace = True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class DDNet(nn.Module):
    '''
    Designed for 112x112 input, keep higher feature map:
    the first conv is set to be (kernel=3, stride=1),
    remove maxpooling in the first stage.
    fewer channels [64, 128, 256, 512].
    '''
    def __init__(self, block, layers, feature_dim, final_reduction_dim=0, channels=[64,128,256,512], checkpoints=[0,0,0,0],
                 ratios=[1,1,1,1], groups=[1,1,1,1],
                 act_type='prelu', fc_type='EMom', use_se=False,
                 bn_mom=0.1, bn_eps=2e-5, bypass_last_bn=False,
                 att_mode='none'):

        super(DDNet, self).__init__()
        self.act_type = act_type
        self.inplanes = channels[0]
        self.feature_dim = feature_dim
        self.checkpoints = checkpoints
        self.use_se = use_se
        self.att_mode = att_mode
        self.final_reduction_dim = final_reduction_dim

        global BN_q, bypass_bn_weight_list
        def BNFunc(*args, **kwargs):
            return nn.BatchNorm2d(*args, **kwargs, momentum=bn_mom, eps=bn_eps)
        BN_q = BNFunc

        bypass_bn_weight_list = []

        m = OrderedDict()
        m['conv1'] = nn.Conv2d(3, self.inplanes, kernel_size=3, stride=1, padding=1, bias=False)
        m['bn1'] = BN_q(self.inplanes)
        m['act1'] = activation(act_type)
        self.group1 = nn.Sequential(m)

        if self.att_mode == 'none':
            self.layer1 = self._make_layer(block, channels[0], layers[0], stride=2, checkpoints=checkpoints[0], ratio=ratios[0], groups=groups[0])
            self.layer2 = self._make_layer(block, channels[1], layers[1], stride=2, checkpoints=checkpoints[1], ratio=ratios[1], groups=groups[1])
            self.layer3 = self._make_layer(block, channels[2], layers[2], stride=2, checkpoints=checkpoints[2], ratio=ratios[2], groups=groups[2])
            self.layer4 = self._make_layer(block, channels[3], layers[3], stride=2, checkpoints=checkpoints[3], ratio=ratios[3], groups=groups[3])

        if self.final_reduction_dim > 0:
            m = OrderedDict()
            m['bn1'] = BN_q(channels[3]*block.expansion)
            m['conv1'] = conv1x1(channels[3]*block.expansion, self.final_reduction_dim, stride=1)
            self.final_reduction = nn.Sequential(m)
            self.fc1 = head.__dict__['get_fc_'+fc_type](self.final_reduction_dim, 7, 7, feature_dim)
        else:
            self.fc1 = nn.Sequential(nn.AdaptiveAvgPool2d(1),nn.Conv2d(channels[3]*block.expansion,1,kernel_size=1,stride=1,padding=0,bias=False))

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                # init.xavier_normal(m.weight.data)
                fan_in = m.out_channels * m.kernel_size[0] * m.kernel_size[1]
                scale = math.sqrt(2. / fan_in)
                m.weight.data.uniform_(-scale, scale)
                if m.bias is not None:
                    m.bias.data.zero_()
            # Xavier can not be applied to less than 2D.
            elif isinstance(m, nn.BatchNorm2d):
                if not m.weight is None:
                    m.weight.data.fill_(1)
                if not m.bias is None:
                    m.bias.data.zero_()
            elif isinstance(m, nn.Linear):
                scale = math.sqrt(3. / m.in_features)
                m.weight.data.uniform_(-scale, scale)
                if m.bias is not None:
                    m.bias.data.uniform_(-scale, scale)

        if bypass_last_bn:
            assert block is BasicBlock_v3
            for param in bypass_bn_weight_list:
                param.data.zero_()
            print('bypass {} bn.weight in {}'.format(len(bypass_bn_weight_list), block))


    def _make_layer(self, block, planes, blocks, stride=1, checkpoints=0, ratio=1, groups=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                BN_q(planes * block.expansion),
            )

        layers = []

        layers.append(block(self.inplanes, planes, stride, downsample, use_se=self.use_se,
                                act_type=self.act_type, use_checkpoint=checkpoints>0))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, use_se=self.use_se,
                                    act_type=self.act_type, use_checkpoint=checkpoints>i))

        return nn.Sequential(*layers)

    def forward(self, x):
        output = {''}

        img_list = []
        # import pdb; pdb.set_trace()
        for cnt in range(x.size(0)):
            tmp = x[cnt]
            tmp = tmp.cpu().numpy()
            tmp = tmp.astype(np.uint8)
            tmp = tmp.transpose((1, 2, 0))
            tmp = cv2.resize(tmp[1:, 1:, :], (112, 112))
            tmp = tmp*3.2/255.0 - 1.6
            tmp = tmp.transpose((2, 0, 1))
            tmp = torch.from_numpy(tmp)
            tmp = tmp.float()
            tmp = tmp[None, ...]
            img_list.append(tmp)
        x = torch.cat(img_list, 0)
        x = x.cuda()

        ori = x
        # import pdb; pdb.set_trace()
        x = self.group1(x)

        if self.att_mode == 'none':
            x = self.layer1(x)
            x = self.layer2(x)
            x = self.layer3(x)
            x = self.layer4(x)
        else:
            raise RuntimeError('unknown att_mode: {}'.format(self.att_mode))

        if self.final_reduction_dim > 0:
            x = self.final_reduction(x)
        headout = self.fc1(x)
        output = {'D-Score':headout}

        return output

class MyModel(object):
    def __init__(self, feature_dim, ckpt_file):
        model_ddnet = DDNet(BasicBlock_v3, [2, 2, 2, 2], feature_dim=feature_dim, channels=[8,16,32,48])
        self.model_ddnet = model_ddnet.cuda()

        cudnn.benchmark = True

        def map_func_cuda(storage, location):
            return storage.cuda()
        ckpt = torch.load(ckpt_file, map_location=map_func_cuda)
        state_dict = remove_prefix(ckpt['state_dict'], 'module.base.')
        self.model_ddnet.load_state_dict(state_dict, strict=True)


    def eval(self):
        self.model_ddnet.eval()

    def __call__(self, input_data):
        
        d_score = self.model_ddnet(input_data.cuda())
        d_score= d_score['D-Score'].data.cpu().numpy()

        return d_score
